/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

/*  LWPR is a regresion technique mainly developed by Sethu
    Vijayakumar. This file tests an implementation of this algorithm,
    coded by Narayanan Edakunni and Giorgos Petkos, cleaned up and
    made more efficient by myself.

    When using this code, please cite

    Sethu Vijayakumar, Aaron D'Souza and Stefan Schaal, "Incremental Online 
    Learning in High Dimensions", Neural Computation, 17:12, pp. 2602-2632 
    (2005) */

#ifndef GP_LWPR_H
#define GP_LWPR_H

#include "array.h"

/*const double initD=1.0;
const double init_lambda = 0.999;
const double tau_lambda = 0.99999;
const double final_lambda = 0.9999;
const double add_threshold = 0.5;
const bool meta = false;
const double meta_rate = 25;
*/

/*! This class defines the local models used by Lwpr - for experts
  only. Please refer to the Lwpr class first. */
class Lmodel {
public:
  double initD;//=1.0;
  double init_lambda;// = 0.999;
  double tau_lambda;// = 0.99999;
  double final_lambda;// = 0.9999;
  double add_threshold;// = 0.5;
  bool meta;// = false;
  double meta_rate;// = 25;
  
  doubleA center;  //!< center of the model
  doubleA mean_x;  //!< x00
  double mean_y;   //!<beta00
  doubleA  W;	   //!< running sum of weights
  double MyMSE_R;  //!< don't know
  doubleA  lambda; //!< forgetting parameter
  doubleA  M;	   //!< cholesky decomposition of dist
  doubleA  n_data; //!< number of data seen

  
  MT::Array<doubleA> uprojections;   //!<u projections
  MT::Array<doubleA> pprojections;   //!<p projections
  MT::Array<doubleA> sXresYres;
  doubleA betas;   //!< betas
  doubleA MSE;     //!< squared errors

  doubleA azz;
  doubleA azres;
  MT::Array<doubleA> axz;
  
  //distance update statistics
  double sum_e2;
  doubleA  sum_ecv2;
  doubleA  aH;
  doubleA  aG;
  double aE;
  
  //penalty
  double gamma;
  //learning param
  doubleA  alpha;
  
  doubleA  b;
  doubleA  h;
  
  double apk;
  doubleA  dist;

public:
  int R;   //!< number of projections
  bool trustworthy;
  bool degenerated; //!< used for pruning of numerically degenerated local models
  
public:
  Lmodel(){};
  Lmodel(const doubleA& x, double y);
  Lmodel(const Lmodel& model, const doubleA& x, double y);
  double activation(const doubleA& x);
  void updateMean(const doubleA& x, double y);
  void updateStat(double w);
  void updateError(const doubleA& x, double y);
  void updateError_matlab(double, double); // emulates the error measure that's used in Matlab's LWPR implementation
  void update(const doubleA& x, double y);
  double predict(const doubleA& x ) const;
  void  update_dist(const doubleA& x, double y, double e_cv, double e);
  double gete(const doubleA& x, double y);
  double gete_cv(const doubleA& x, double y);
  bool check_add_projection(const doubleA& x, double y);
  void add_projection(const doubleA& x, double y);
  void printProj();
  friend std::ostream& operator<<(std::ostream& outs, const Lmodel& lm);

  void save(std::ostream&);
  void load(std::istream&);
  
  double s2pred(const doubleA& x, double w);
  void updateApk(const doubleA& x,double w);
  const doubleA& getCenter() const{ return center; }
  
private :
  void computeZs(doubleA& zs, const doubleA& x);
  void loadParameters(uint N);
  void initialise(int N);
  //void reinitialize();
  doubleA convert_vec(const doubleA& x);
  void getdJdM(doubleA& dJdM,const doubleA& D,const doubleA& M,
    const doubleA& x, const doubleA& center,
    const doubleA& z, const doubleA& azz, double w,
    double W, double e_cv, double e, double gamma,
    const doubleA& derivative_ok, double& sum_dJ1dw,
    doubleA& dwdM);
  
  double getsum_dJ1dw(double e_cv, double e, double w, double W,
    const doubleA& z_vec, const doubleA& azz_vec,
    const doubleA& derivative_ok);
  void update_dist_stat(double e_cv, double w, const doubleA& z_vec,
    const doubleA& azz_vec, double transient_multiplier,
    const doubleA& derivative_ok);
  void getdJ2dM(doubleA& dJ2dM,const doubleA& D, const doubleA& M, double gamma);
  void getdwdM(doubleA& dwdM,const doubleA& M, const doubleA& x, 
    const doubleA& center, double w);
  void getdDdMkl(doubleA& dDdMkl,const doubleA& M, int k, int l);
  
  //meta methods
  void meta_update(double e_cv, double W, double w, double e, const doubleA& aH,
    double aE, const doubleA& z_vec, 
    const doubleA& azz_vec, const doubleA& derivative_ok,
    const doubleA& x, const doubleA& center, double dJ1dw,
    const doubleA& dwdM, double transient_multiplier, const doubleA& dJdM);
  void getdJ2dJ2dMdM(doubleA& dJ2dJ2dMdM,const doubleA& D, const doubleA& M, 
    double gamma);
  void getdwdwdMdM(doubleA& dwdwdMdM,const doubleA& M, const doubleA& x, 
    const doubleA& center, double w);
  
  void cholesky(doubleA& L,const doubleA& A);
  intA find(const doubleA& A, double value, bool gt);
  double absmax(const doubleA& A);
};

/*! LWPR is a regresion technique mainly developed by Sethu
    Vijayakumar. This file tests an implementation of this algorithm,
    coded by Narayanan Edakunni and Giorgos Petkos, cleaned up and
    made more efficient by myself.

    When using this code, please cite

    Sethu Vijayakumar, Aaron D'Souza and Stefan Schaal, "Incremental Online 
    Learning in High Dimensions", Neural Computation, 17:12, pp. 2602-2632 
    (2005) */
class Lwpr {
public:
  //! Constructor, takes parameters from parameter file.
  Lwpr();

  //! updates the LWPR model given a new datapoint x->y, or a batch of data
  void learn(const doubleA& X,const doubleA& Y);

  //! predicts the output Y for a given input X -- also for batch data
  void map(const doubleA& X,doubleA& Y);
  
  //! same as learn(X,Y) - for compatibility reasons
  void updaterfs(const doubleA& X,const doubleA& Y){ learn(X,Y); }

  //! same { doubleA y; map(x,y); return y; } - for compatibility reasons
  doubleA predict(const doubleA& x){ doubleA y; map(x,y); return y; }

  /*!\brief tells LWPR to use a specific normalization for inputs (during both, learn and map):
      the mean is subtracted, the trans multiplied (trans can be 0-, 1-, or 2-dimensional
      for scalar, vector-element-wise and matrix multiplications, respectively) */
  void useNormalization(const doubleA& mean,const doubleA& trans){
    normMean=mean; normTrans=trans;
  }

  /*!\brief uses a normalization learned from the given batch of input data:
      the i-th component of an input vector will be multiplied by
      1./standard-deviation(i-th dimension) */
  void learnNormalization(doubleA& X);

  //! saves the LWPR model in the file given as parameter
  void save(char *);

  //! loads the LWPR model from the file given as parameter
  void load(char *,bool alsoParameters=true);

  //! get the confidence bounds (standard deviation) at some input point
  double confidence(const doubleA &x){
    doubleA nx;
    normalizeInput(nx,x);
    return ::sqrt(s2pred(nx));
    //@todo allows for negative prediction confidences to exist (pathological cases)
    //return ( s2 >= 0 ? s2 : -s2 ); // this would swap sign in pathological cases
  }
  
  //! preliminary -- output useful readable state information
  void report(std::ostream& os){
    uint i,j;
    os
      <<"LWPR report:"
      <<"\n inDim="<<in_dim <<" outDim="<<out_dim <<std::endl;
    for(i=0;i<models.N;i++){
      os <<i <<"th dimension: #models=" <<models(i).N <<std::endl;
      for(j=0;j<models(i).N;j++){
	os
	  <<' ' <<j <<"th model: center="
	  <<((models(i)(j)).center)
	  <<" mean="
	  <<((models(i)(j)).mean_x)
	  <<std::endl;
      }
    }
  }
  
  //! get the number of rfs for the given output dimension
  int get_rfs_no(int out_dim);

  //! average number of projection dimensions used in the local models
  double get_proj_average(int out_dim);

  uint rfsno; //!< number of receptive fields

  //normalizations
  doubleA normMean,normTrans;

private:
  void initialize(uint inDim,uint outDim);
  double s2pred(const doubleA& x);
  double predict(const MT::Array<Lmodel>& rfs, const doubleA& x);
  void updaterfs(MT::Array<Lmodel>& rfs, const doubleA& x, double y);
  void normalizeInput(doubleA& normIn,const doubleA& unNormIn);


  bool isInitialized;
  uint in_dim;
  uint out_dim;		
  bool add_proj;
  bool updateD;
  double cut_off;
  double w_prune;
  double w_gen;

  MT::Array<MT::Array<Lmodel> > models;  //main data structure!
};

#ifdef MT_IMPLEMENTATION
#include "lwpr.cpp"
#endif

#endif
